import pandas as pd
import logging
from SPARQLWrapper import SPARQLWrapper, DIGEST, JSON, XML, TURTLE, N3, RDF
logging.basicConfig(filename='drupal.log',level=logging.DEBUG)

def getLabel(uri):
	sparql = SPARQLWrapper("https://carre.kmi.open.ac.uk/sparql")
	sql = """SELECT ?o WHERE {
	  { <"""+uri+"""> <http://www.w3.org/2000/01/rdf-schema#label> ?o .}
	}"""
	sparql.setQuery(sql)
	sparql.setReturnFormat(JSON)
	results = sparql.query().convert()
	for record in results['results']['bindings']:
		return(record['o']['value'])

	return None


def describe(uri, ignoreOther=True):
	sparql = SPARQLWrapper("https://carre.kmi.open.ac.uk/sparql")
	sql = """SELECT ?p ?o WHERE {
	  { <"""+uri+"""> ?p ?o .}
	}"""
	sparql.setQuery(sql)
	sparql.setReturnFormat(JSON)
	results = sparql.query().convert()
	rs = []
	for record in results['results']['bindings']:
		if ignoreOther:
			p = record['p']['value']
			if (not p.startswith("http://carre.kmi.open.ac.uk/ontology/risk.owl")) and (p!="http://www.w3.org/1999/02/22-rdf-syntax-ns#type") and (p!="http://purl.org/dc/terms/created") and (p!="http://purl.org/dc/terms/modified"):
				continue
		o = record['o']['value']
		if 'datatype' in record['o'].keys():
			o = "\"" + o + "\"^^" + record['o']['datatype']
		# else:
		# 	print record['o']
		# print o
		if o.startswith("https://carre.kmi.open.ac.uk/taxonomy_term"):
			o = getLabel(o)
		rs.append((uri, record['p']['value'], o))

	return rs

def fetchInstances(type="observable"):
	"""type: observable, risk_element, risk_factor, risk_evidence, citation"""
	type = "http://carre.kmi.open.ac.uk/ontology/risk.owl#" + type
	sparql = SPARQLWrapper("https://carre.kmi.open.ac.uk/sparql")
	sql = """SELECT distinct ?s WHERE {
	  { ?s <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <"""+type+"""> . 
	  ?s ?p ?o .}
	}"""
	sparql.setQuery(sql)
	sparql.setReturnFormat(JSON)
	results = sparql.query().convert()
	rs = []
	for record in results['results']['bindings']:
		rs.append(record['s']['value'])

	return rs

def extractAll():
	observables = fetchInstances("observable")
	rs = []
	for observable in observables:
		print "fetching observable", observable
		rs = rs + describe(observable)

	elements = fetchInstances("risk_element")
	for element in elements:
		print "fetching element", element
		rs = rs + describe(element)

	risk_factors = fetchInstances("risk_factor")
	for risk_factor in risk_factors:
		print "fetching risk_factor", risk_factor
		rs = rs + describe(risk_factor)

	risk_evidences = fetchInstances("risk_evidence")
	for risk_evidence in risk_evidences:
		print "fetching risk_evidence", risk_evidence
		rs = rs + describe(risk_evidence)

	citations = fetchInstances("citation")
	for citation in citations:
		print "fetching citation", citation
		rs = rs + describe(citation)

	df = pd.DataFrame(rs)
	df.columns = ['subject', 'predicate', 'object']
	return df

def getCodeURL(st):
	if st is None:
		return None
	tmp = st.lower()
	rs = tmp.split(":")
	if len(rs)>1:
		code = rs[1].strip()
	else:
		try:
			if tmp.startswith("c"):
				i = int(tmp[1:])
				code = "C"+tmp[1:]
				tmp = "umls_cui: " + tmp[1:]
		except Exception, e:
			return None

	if tmp.startswith("http://"):
		if tmp.startswith("http://purl.bioontology.org/ontology/"):
			return "http://umls.nlm.nih.gov/sab/"+st.split("/")[-2].strip() + "/cui/" + st.split("/")[-1].strip()
		else:
			return tmp
	elif "umls_cui" in tmp:
		return "http://umls.nlm.nih.gov/sab/MTH/cui/"+ code.upper()
	elif "snowmed_cui" in tmp:
		return "http://umls.nlm.nih.gov/sab/SNOWMED/cui/" + code.upper()
	elif "chebi" in tmp:
		return "http://purl.obolibrary.org/obo/CHEBI_" + code.upper()
	return None

def capitaliseFirstLetter(st):
	if not st.startswith("http://carre.kmi.open.ac.uk/ontology/risk.owl#"):
		return st
	else:
		tmp = st.replace("http://carre.kmi.open.ac.uk/ontology/risk.owl#", "")
		return "http://carre.kmi.open.ac.uk/ontology/risk.owl#" + tmp[0].capitalize()+ tmp[1:]


def fixObjects(df):

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#has_author						-> uri
	df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_author'] = "http://carre.kmi.open.ac/authors/"+df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_author'].map(removeWhiteSpace)

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#has_reviewer 						-> uri
	df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_reviewer'] = "http://carre.kmi.open.ac/authors/"+df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_reviewer'].map(removeWhiteSpace)

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_datatype 						-> uri: 'Plain text', u'Other', u'Integer', u'Real'
	df.object[(df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_datatype')] = df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_type'].apply(removeWhiteSpace)

	df.object[(df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_factor_association_type")] = df.object[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_factor_association_type"].apply(capitaliseFirstLetter)

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_author 							-> uri
	df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_author'] = "http://carre.kmi.open.ac/authors/"+df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_author'].map(removeWhiteSpace)

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_identifier							-> external ontology/string
	df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_identifier'] = df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_identifier'].map(getCodeURL)

	df.object[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_evidence_ratio_type"] = df.object[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_evidence_ratio_type"].str.strip()

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_unit							-> external ontology/string
	logging.warning("[observable_unit] contains unidentified observable units, see " + " ".join(set(df.subject[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_unit'].tolist())))
	# df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_unit'] = "\"" + df.object[df.predicate=='https://carre.kmi.open.ac.uk/ontology/risk.owl#observable_unit'] + "\"^^xsd:string"

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_name							-> null
	df = df[df.predicate!='http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_name']

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_identifier							-> external ontology/string
	df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_identifier'] = df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_identifier'].map(getCodeURL)

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_type							-> biomedical (including condition, disorder and/or disease)|biomedical (including condition, disorder and/or disease)
																									# demographic (age, sex, race, occupation, NBI, education, etc)|demographic (age, sex, race, occupation, NBI, education, etc)
																									# genetic (genetic polymorphism)|genetic (genetic polymorphism)
																									# behavioural (physical activity, diet, smoking, etc.)|behavioural (physical activity, diet, smoking, etc.)
																									# intervention (drugs, etc)|intervention (drugs, etc)
																									# environmental (air pollution, water contamination)|environmental (air pollution, water contamination)
	df.object[(df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_type')] = df.object[(df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_type')].str.strip()

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_modifiable_status					-> xsd:Boolean
	df.object[(df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_modifiable_status')&(df.object=="1")] = "\"true\"^^xsd:boolean"
	df.object[(df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_modifiable_status')&(df.object=="")] = "\"false\"^^xsd:boolean"
	
	# http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_condition
	df.object[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_condition"] = "\"\"\"" + df.object[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_condition"] + "\"\"\"^^xsd:string"

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_factor_title							-> string
	df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_factor_title'] = "\"" + df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_factor_title'] + "\"^^xsd:string"

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#is_adjusted_for						-> taxonomy/string
	logging.warning("[is_adjusted_for] contains unidentified entities, see " + " ".join(set(df.subject[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#is_adjusted_for"].tolist())))
	df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#is_adjusted_for'] = "\"" + df.object[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#is_adjusted_for'] + "\"^^xsd:string"

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#has_citation_source_type						-> taxonomy/string
	logging.warning("[has_citation_source_type] contains unidentified entities, see " + " ".join(set(df.subject[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_citation_source_type"].tolist())))

	return df


def replaceURIs(df, to_replace, replace_with):
	df.subject[df.subject==to_replace] = replace_with
	df.object[df.object==to_replace] = replace_with
	return df

def titleToURI(title):
	title = title.replace("^^xsd:string", "")
	title = title.replace("^^xsd:int", "")
	title = title.replace("\"", "")
	title = title.strip()
	title = title.replace(" ","_")
	title = title.replace(":","")
	title = title.lower()
	return title

def removeWhiteSpace(st):
	if st is None:
		return ""
	st = st.replace(" ","_")
	return st

def removeBreakLines(st):
	st = st.replace("\n","\\\\\n")
	return st

def fixSubjects(df):
	# observables -> observables/<title>
	# risk_element ->risk_elements/<title>
	# risk_factor -> risk_factors/<title>
	# risk_evidence -> risk_evidence/<drupal_node_id>
	# citation -> citations/<pbmid>
	subjects = df.subject.unique()
	d = {}
	for subject in subjects:
		type = df[(df.subject==subject)&(df.predicate=="http://www.w3.org/1999/02/22-rdf-syntax-ns#type")].object.tolist()[0]
		d[subject] = type
	
	for subject in d.keys():
		if d[subject]=="http://carre.kmi.open.ac.uk/ontology/risk.owl#citation":
			rs = df[(df.subject==subject)&(df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_citation_pubmed_identifier")].object.tolist()
			if len(rs)<1:
				continue
			else:
				pbmid = titleToURI(rs[0])
			df = replaceURIs(df, subject, "http://carre.kmi.open.ac.uk/citations/"+pbmid)
		elif d[subject]=="http://carre.kmi.open.ac.uk/ontology/risk.owl#observable":
			title = df[(df.subject==subject)&(df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_observable_name")].object.tolist()[0]
			title = titleToURI(title)
			df = replaceURIs(df, subject, "http://carre.kmi.open.ac.uk/observables/"+title)
		elif d[subject]=="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element":
			title = df[(df.subject==subject)&(df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#has_risk_element_name")].object.tolist()[0]
			title = titleToURI(title)
			df = replaceURIs(df, subject, "http://carre.kmi.open.ac.uk/elements/"+title)
		elif d[subject]=="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence":
			title = "http://carre.kmi.open.ac.uk/CARRE_RE" + subject.split("/")[-1]
			df = replaceURIs(df, subject, title)
		elif d[subject]=="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_factor":
			title = "http://carre.kmi.open.ac.uk/CARRE_RF" + subject.split("/")[-1]
			df = replaceURIs(df, subject, title)
		elif d[subject]=="http://carre.kmi.open.ac.uk/ontology/risk.owl#confidence_interval":
			title = "http://carre.kmi.open.ac.uk/CARRE_CI" + subject.split("/")[-1]
			df = replaceURIs(df, subject, title)
		# df.replace(to_replace=subject,value=title,inplace=True)
	
	return df

def fixPredicates(df):

	# http://www.w3.org/1999/02/22-rdf-syntax-ns#type
	#nothing

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_minimum_confidence_interval
	for subject in df[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_minimum_confidence_interval"].subject.unique():
		d = {}
		d['subject'] = subject
		d['predicate'] = "http://carre.kmi.open.ac.uk/ontology/risk.owl#has_confidence_interval"
		d['object'] = subject + "_interval"
		df = pd.concat([df, pd.DataFrame(d, index={0})])
		d = {}
		d['subject'] = subject+"_interval"
		d['predicate'] = "http://carre.kmi.open.ac.uk/ontology/risk.owl#has_minimum_confidence_interval"
		d['object'] = df[(df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_minimum_confidence_interval")&(df.subject==subject)].object.tolist()[0]
		df = pd.concat([df, pd.DataFrame(d, index={0})])
		d = {}
		d['subject'] = subject+"_interval"
		d['predicate'] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
		d['object'] = "http://carre.kmi.open.ac.uk/ontology/risk.owl#confidence_interval"
		df = pd.concat([df, pd.DataFrame(d, index={0})])
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_minimum_confidence_interval"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_maximum_confidence_interval
	for subject in df[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_maximum_confidence_interval"].subject.unique():
		d = {}
		d['subject'] = subject+"_interval"
		d['predicate'] = "http://carre.kmi.open.ac.uk/ontology/risk.owl#has_maximum_confidence_interval"
		d['object'] = df[(df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_maximum_confidence_interval")&(df.subject==subject)].object.tolist()[0]
		df = pd.concat([df, pd.DataFrame(d, index={0})])	
		d = {}
		d['subject'] = subject+"_interval"
		d['predicate'] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
		d['object'] = "http://carre.kmi.open.ac.uk/ontology/risk.owl#confidence_interval"
		df = pd.concat([df, pd.DataFrame(d, index={0})])
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_maximum_confidence_interval"]	

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_title
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_evidence_title"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_unit
	# todo: fix this with lookup

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_datatype
	# todo: fix this with lookup

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_type
	df.predicate[df.predicate=="http://carre.kmi.open.ac.uk/ontology/risk.owl#observable_type"] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_title
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_title"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_factor_title
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_factor_title"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_name
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_name"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_type
	df.predicate[df.predicate=='http://carre.kmi.open.ac.uk/ontology/risk.owl#risk_element_type'] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_journal
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_journal"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_ocebm_level
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_ocebm_level"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_authors
	df = df[df.predicate!="http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_authors"]

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_year
	df = df[df.predicate!='http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_year']

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_volume
	df = df[df.predicate!='http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_volume']

	# http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_issue
	df = df[df.predicate!='http://carre.kmi.open.ac.uk/ontology/risk.owl#citation_issue']

	return df

def fixAll(df):
	import numpy as np
	df = fixObjects(df)
	df = fixPredicates(df)
	df = fixSubjects(df)
	df = df.replace("\"\"^^xsd:string", np.nan)
	df = df.replace("\"\"^^xsd:int", np.nan)
	df = df.replace("\"\"^^xsd:float", np.nan)
	df = df.replace("none", np.nan)
	df = df.replace("N/A", np.nan)
	df = df.replace("-", np.nan)
	df = df[df.predicate!="http://purl.org/dc/terms/created"]
	df = df[df.predicate!="http://purl.org/dc/terms/modified"]
	tmp = df.fillna(0)
	logging.warning("****SUBJECT - PREDICATE****")
	for tuple in tmp[tmp.object==0][['subject','predicate']].itertuples():
		logging.warning(tuple[1] + " - " + tuple[2])
	df = df.dropna()
	df = df[df.object!="\"\"^^xsd:integer"]
	return df

def prepareTriples(df):
	triples = []

	for tuple in df.itertuples():
		if isinstance(tuple[1], float):
			print "skipping", tuple[2], tuple[1]
			continue
		subject = tuple[3]
		predicate = tuple[2]
		object = tuple[1]
		if object is None:
			print "object is None"
			continue
		if subject.startswith("http://"):
			subject = "<" + subject + ">"
		if predicate.startswith("http://"):
			predicate = "<" + predicate + ">"
		if object.startswith("http://"):
			object = "<" + object + ">"
		triples.append(subject + " " + predicate + " " + object + " .")

	return triples

def insertTriples(triples, bufferSize=5):
	from SPARQLWrapper import SPARQLWrapper, DIGEST, JSON, XML, TURTLE, N3, RDF
	import numpy as np

	divider = int(len(triples)/float(bufferSize))
	batches = np.array_split(triples, divider)
	sparql = SPARQLWrapper("https://carre.kmi.open.ac.uk/sparql-auth")
	sparql.setHTTPAuth(DIGEST)
	sparql.setCredentials("USERNAME", "PASSWORD")
	i = 0
	for batch in batches:
		i += 1
		st = "\n".join(list(batch))
		sql = "INSERT IN <http://carre.kmi.open.ac.uk/beta> {\n " + st + "  }"
		# print "\n\n",sql
		sparql.setQuery(sql)
		results = sparql.query().convert()
		print "finished", i, "of", len(batches)

	return